
#include "Trailblazer.h"
#include "random.h"
#include "map.h"
using namespace std;

static const double SUFFICIENT_DIFFERENCE = 0.2;

/**
anton-search():

    create an empty path

    make a current node equal to the start node
    color the start node green
    add the start node to the path

    while (the current node is not the end node) {
        randomly sample a new current node that is a neighbor of current
        if (current node has been seen more than N times) {
            return an empty path
        }
        color current green
        add current to the path
    }


    return the constructed path
*/

RoadNode* getRandomNeighbor(const RoadGraph& graph, RoadNode* node) {
    Set<RoadNode*> neighborSet = graph.neighborsOf(node);
    int sizeOfSet = neighborSet.size();
    if (sizeOfSet == 0) return node;
    int randomInt = randomInteger(0, sizeOfSet - 1);
    int counter = 0;
    RoadNode* neighborToReturn;
    for (RoadNode* neighbor : neighborSet) {
        if (counter == randomInt)  {
            neighborToReturn = neighbor;
            break;
        }
        counter++;
    }
    return neighborToReturn;
}

// Threshold of failure
static const int N = 50;


Path antonSearch(const RoadGraph& graph, RoadNode* start, RoadNode* end) {
    Path path;


    Map<RoadNode*, int> visits;

    RoadNode* current = start;
    start->setColor(Color::GREEN);
    path.add(start);

    while (current != end) {
        visits[current]++;
        current = getRandomNeighbor(graph, current);
        if (visits[current] > N) return {};
        current->setColor(Color::GREEN);
        path.add(current);
    }

    return path;
}


Path breadthFirstSearch(const RoadGraph& graph, RoadNode* start, RoadNode* end) {
    return antonSearch(graph, start, end);
}

Path dijkstrasAlgorithm(const RoadGraph& graph, RoadNode* start, RoadNode* end) {
    /* TODO: Delete the following lines and implement this function! */
    (void) graph;
    (void) start;
    (void) end;
    return {};
}

Path aStar(const RoadGraph& graph, RoadNode* start, RoadNode* end) {
    /* TODO: Delete the following lines and implement this function! */
    (void) graph;
    (void) start;
    (void) end;
    return {};
}

Path alternativeRoute(const RoadGraph& graph, RoadNode* start, RoadNode* end) {
    /* TODO: Delete the following lines and implement this function! */
    (void) graph;
    (void) start;
    (void) end;
    return {};
}

